/* 
 * UcxSockChan.c
 *
 *	UCX driver for channels based on TCP sockets.
 *
 */

#include "tclInt.h"		/* Internal definitions for Tcl. */
#include "tclPort.h"		/* Portability features for Tcl. */

#include <stdio.h>
#include <starlet.h>
#include <ioctl.h>
#include <inetiodef.h>
#include <errno.h>

/*
 * This structure describes per-instance state of a tcp based channel.
 */

typedef struct TcpState {
    int flags;				/* ORed combination of the
                                         * bitfields defined below. */
    Tcl_File sock;			/* The socket itself. */
    Tcl_TcpAcceptProc *acceptProc;	/* Proc to call on accept. */
    ClientData acceptProcData;		/* The data for the accept proc. */

    int chan;
    int eventFlag;
    int checkMask;
    int watchMask;
    short iosb[4];
} TcpState;

/*
 * These bits may be ORed together into the "flags" field of a TcpState
 * structure.
 */

#define TCP_ASYNC_SOCKET	(1<<0)
#define TCP_ASYNC_CONNECT	(1<<1)
#define TCP_WATCHING_SOCKET	(1<<2)

/*
 * Static routines for this file:
 */

static TcpState *	CreateSocket _ANSI_ARGS_((Tcl_Interp *interp,
			    int port, char *host, int server,
			    char *myaddr, int myport));
static int		CreateSocketAddress _ANSI_ARGS_(
			    (struct sockaddr_in *sockaddrPtr,
			    char *host, int port));
static void		TcpAccept _ANSI_ARGS_((ClientData data, int mask));
static int		TcpBlockMode _ANSI_ARGS_((ClientData data,
		            Tcl_File inFile, Tcl_File outFile, int mode));
static int		TcpClose _ANSI_ARGS_((ClientData instanceData,
			    Tcl_Interp *interp, Tcl_File inFile,
			    Tcl_File outFile));
static int		TcpInput _ANSI_ARGS_((ClientData instanceData,
		            Tcl_File infile, char *buf, int toRead,
        		    int *errorCode));
static int		TcpOutput _ANSI_ARGS_((ClientData instanceData,
		            Tcl_File outFile, char *buf, int toWrite,
		            int *errorCode));

static void		socketReadyHandler(TcpState *);
static void		socketFreeProc(ClientData);

/*
 * This structure describes the channel type structure for TCP socket
 * based IO:
 */

static Tcl_ChannelType tcpChannelType = {
    "tcp",				/* Type name. */
    TcpBlockMode,			/* Set blocking/nonblocking mode.*/
    TcpClose,				/* Close proc. */
    TcpInput,				/* Input proc. */
    TcpOutput,				/* Output proc. */
    NULL,				/* Seek proc. */
    NULL,				/* Set option proc. */
    NULL				/* Get option proc. */
};

/*
 * The preferred buffer size for Unix channels.
 */

#define SOCKET_BUFSIZE	4096


/*
 *----------------------------------------------------------------------
 *
 * TcpBlockMode --
 *
 *	Set blocking or non-blocking mode on a TCP channel.
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpBlockMode(instanceData, inFile, outFile, mode)
    ClientData instanceData;		/* Socket state. */
    Tcl_File inFile, outFile;		/* Input, output for channel. */
    int mode;				/* The mode to set. */
{
    unsigned int enable;
    int sock;
    TcpState *statePtr;
    
    statePtr = (TcpState *) instanceData;
    if (mode == TCL_MODE_BLOCKING) {
        statePtr->flags &= (~(TCP_ASYNC_SOCKET));
    } else {
        statePtr->flags |= TCP_ASYNC_SOCKET;
    }
    enable = (mode == TCL_MODE_BLOCKING) ? 0 : 1;
    if (inFile != NULL) {
        sock = (int) Tcl_GetFileInfo(inFile, NULL);
	if (socket_ioctl(sock, FIONBIO, &enable) < 0)
	    return socket_errno;
    }
    if (outFile != NULL) {
        sock = (int) Tcl_GetFileInfo(outFile, NULL);
	if (socket_ioctl(sock, FIONBIO, &enable) < 0)
	    return socket_errno;
    }

    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpInput --
 *
 *	Reads input from the TCP IO channel into the buffer given. Returns
 *	count of how many bytes were actually read, and an error indication.
 *
 * Results:
 *	A count of how many bytes were read is returned and an error
 *	indication is returned in an output argument.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpInput(instanceData, inFile, buf, bufSize, errorCode)
    ClientData instanceData;		/* Socket state. */
    Tcl_File inFile;			/* Input device for channel. */
    char *buf;				/* Where to store data read. */
    int bufSize;			/* How much space is available
                                         * in the buffer? */
    int *errorCode;			/* Where to store error code. */
{
    TcpState *statePtr;			/* The state of the socket. */
    int sock;				/* The OS handle. */
    int bytesRead;			/* How many bytes were read? */
    int state;				/* Of waiting for connection. */
    int timeOut;			/* Of waiting for connection. */

    *errorCode = 0;
    socket_errno = 0;
    sock = (int) Tcl_GetFileInfo(inFile, NULL);
    statePtr = (TcpState *) instanceData;
    
    if (bufSize == 0) {
        return 0;
    }

    /*
     * If an asynchronous connect is in progress, attempt to wait for it
     * to complete before reading.
     */
    
    if (statePtr->flags & TCP_ASYNC_CONNECT) {
        if (statePtr->flags & TCP_ASYNC_SOCKET) {
            timeOut = 0;
        } else {
            timeOut = -1;
        }
        state = TclWaitForFile(inFile, TCL_WRITABLE, timeOut);
        if (state & TCL_WRITABLE) {
            statePtr->flags &= (~(TCP_ASYNC_CONNECT));
        } else if (timeOut == 0) {
            *errorCode = socket_errno = EWOULDBLOCK;
            return -1;
        }
    }
            
    bytesRead = socket_read(sock, buf, bufSize);
    if (bytesRead > -1) {
        return bytesRead;
    }
    if (socket_errno == ECONNRESET) {

        /*
         * Turn ECONNRESET into a soft EOF condition.
         */
        
        return 0;
    }
    *errorCode = socket_errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpOutput--
 *
 *	Writes the given output on the TCP IO channel. Returns count of how
 *	many characters were actually written, and an error indication.
 *
 *	NOTE: This routine is needed because on sockets we must use send(2)
 *	instead of write(2) to get reliable error reporting on some Unixes.
 *
 * Results:
 *	A count of how many characters were written is returned and an
 *	error indication is returned in an output argument.
 *
 * Side effects:
 *	Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

static int
TcpOutput(instanceData, outFile, buf, toWrite, errorCode)
    ClientData instanceData;		/* Socket state. */
    Tcl_File outFile;			/* Output device for channel. */
    char *buf;				/* The data buffer. */
    int toWrite;			/* How many bytes to write? */
    int *errorCode;			/* Where to store error code. */
{
    TcpState *statePtr;
    int written;
    int sock;
    int state;				/* Of waiting for connection. */
    int timeOut;			/* Of waiting for connection. */

    *errorCode = 0;
    socket_errno = 0;
    sock = (int) Tcl_GetFileInfo(outFile, NULL);
    statePtr = (TcpState *) instanceData;

    /*
     * If an asynchronous connect is in progress, attempt to wait for it
     * to complete before reading.
     */
    
    if (statePtr->flags & TCP_ASYNC_CONNECT) {
        if (statePtr->flags & TCP_ASYNC_SOCKET) {
            timeOut = 0;
        } else {
            timeOut = -1;
        }
        state = TclWaitForFile(outFile, TCL_WRITABLE, timeOut);
        if (state & TCL_WRITABLE) {
            statePtr->flags &= (~(TCP_ASYNC_CONNECT));
        } else if (timeOut == 0) {
            *errorCode = socket_errno = EWOULDBLOCK;
            return -1;
        }
    }

    written = socket_write(sock, buf, toWrite);
    if (written > -1) {
        return written;
    }
    *errorCode = socket_errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpClose --
 *
 *	Close the socket.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the socket.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpClose(instanceData, interp, inFile, outFile)
    ClientData instanceData;	/* The socket to close. */
    Tcl_Interp *interp;		/* For error reporting - unused. */
    Tcl_File inFile, outFile;	/* Unused. */
{
    TcpState *statePtr;
    Tcl_File sockFile;
    int sock;
    int errorCode = 0;

    statePtr = (TcpState *) instanceData;
    sockFile = statePtr->sock;
    sock = (int) Tcl_GetFileInfo(sockFile, NULL);
    
    /*
     * Delete a file handler that may be active for this socket.
     * Channel handlers are already deleted in the generic IO close
     * code which called this function.
     */
    
    Tcl_DeleteFileHandler(sockFile);

    /*
     * We assume that inFile==outFile==sockFile and so
     * we only clean up sockFile.
     */

    Tcl_FreeFile(sockFile);

    if (socket_close(sock) < 0) {
	errorCode = socket_errno;
    }

    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * CreateSocket --
 *
 *	This function opens a new socket in client or server mode
 *	and initializes the TcpState structure.
 *
 * Results:
 *	Returns a new TcpState, or NULL with an error interp.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static TcpState *
CreateSocket(interp, port, host, server, myaddr, myport)
    Tcl_Interp *interp;		/* For error reporting; can be NULL. */
    int port;			/* Port number to open. */
    char *host;			/* Name of host on which to open port. */
    int server;			/* 1 if socket should be a server socket,
				 * else 0 for a client socket. */
    char *myaddr;		/* Optional client-side address */
    int myport;			/* Optional client-side port */
{
    int enable = 1;
    int status, sock, asyncConnect, curState, origState;
    struct sockaddr_in sockaddr;	/* Socket address */
    struct sockaddr_in mysockaddr;	/* Socket address for client */
    TcpState *statePtr;

    if (! CreateSocketAddress(&sockaddr, host, port)) {
    	goto addressError;
    }
    if ((myaddr != NULL || myport != 0) &&
    	    ! CreateSocketAddress(&mysockaddr, myaddr, myport)) {
	goto addressError;
    }

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	goto addressError;
    }

    /*
     * Set kernel space buffering
     */

    TclSockMinimumBuffers(sock, SOCKET_BUFSIZE);

    asyncConnect = 0;
    if (server) {
	/*
	 * Set up to reuse server addresses automatically and bind to the
	 * specified port.
	 */
    
	status = 1;
	(void) setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &status,
		sizeof(status));
	status = bind(sock, (struct sockaddr *) &sockaddr,
                sizeof(struct sockaddr));
	if (status != -1) {
	    status = listen(sock, 5);
	} 
    } else {
    	if (myaddr != NULL || myport != 0) {
    	    status = 1;
    	    (void) setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &status,
    	    	sizeof(status));
	    status = bind(sock, (struct sockaddr *) &mysockaddr,
	    	    sizeof(struct sockaddr));
	    if (status < 0) {
	    	goto bindError;
	    }
	}

	/*
	 * Attempt to connect. The connect may fail at present with an
	 * EINPROGRESS but at a later time it will complete. The caller
	 * will set up a file handler on the socket if she is interested in
	 * being informed when the connect completes.
	 */

        asyncConnect = 0;
#if 0
        origState = fcntl(sock, F_GETFL);
        curState = origState | O_NONBLOCK;
        status = fcntl(sock, F_SETFL, curState);
#else
	/*
	 * I don't know how to read the current state of a socket.
	 * Assume that it is in blocking mode initially.
	 */
	status = socket_ioctl(sock, FIONBIO, &enable);
#endif
        if (status != -1) {
            status = connect(sock, (struct sockaddr *) &sockaddr,
                    sizeof(sockaddr));
            if (status == -1) {
                if (socket_errno == EINPROGRESS) {
                    asyncConnect = 1;
                    status = 0;
                }
            }
            if (status == 0) {
#if 0
                status = fcntl(sock, F_SETFL, origState);
#else
		enable = 0;
		status = socket_ioctl(sock, FIONBIO, &enable);
#endif
            }
        }
    }

  bindError:
    if (status < 0) {
        if (interp != NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        socket_close(sock);
        return NULL;
    }

    /*
     * Allocate a new TcpState for this socket.
     */

    statePtr = (TcpState *) ckalloc((unsigned) sizeof(TcpState));
    statePtr->flags = 0;
    if (asyncConnect) {
        statePtr->flags = TCP_ASYNC_CONNECT;
    }
    statePtr->sock = Tcl_GetFile((ClientData) sock, TCL_VMS_SOCKET);
    statePtr->chan = sock;
    statePtr->eventFlag = 0;
    statePtr->checkMask = 0;
    statePtr->watchMask = 0;
    Tcl_SetNotifierData(statePtr->sock, socketFreeProc, (ClientData)statePtr);
    
    return statePtr;

  addressError:
    if (interp != NULL) {
	Tcl_AppendResult(interp, "couldn't open socket: ",
		Tcl_PosixError(interp), (char *) NULL);
    }
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * CreateSocketAddress --
 *
 *	This function initializes a sockaddr structure for a host and port.
 *
 * Results:
 *	1 if the host was valid, 0 if the host could not be converted to
 *	an IP address.
 *
 * Side effects:
 *	Fills in the *sockaddrPtr structure.
 *
 *----------------------------------------------------------------------
 */

static int
CreateSocketAddress(sockaddrPtr, host, port)
    struct sockaddr_in *sockaddrPtr;	/* Socket address */
    char *host;				/* Host.  NULL implies INADDR_ANY */
    int port;				/* Port number */
{
    struct hostent *hostent;		/* Host database entry */
    struct in_addr addr;		/* For 64/32 bit madness */

    (void) memset((char *) sockaddrPtr, '\0', sizeof(struct sockaddr_in));
    sockaddrPtr->sin_family = AF_INET;
    sockaddrPtr->sin_port = htons((unsigned short) (port & 0xFFFF));
    if (host == NULL) {
	addr.s_addr = INADDR_ANY;
    } else {
	hostent = gethostbyname(host);
	if (hostent != NULL) {
	    memcpy((char *) &addr,
		(char *) hostent->h_addr_list[0], (size_t) hostent->h_length);
	} else {
	    addr.s_addr = inet_addr(host);
	    if (addr.s_addr == (unsigned long)-1) {
		return 0;	/* error */
	    }
	}
    }
    /*
     * There is a rumor that this assignment may require care on
     * some 64 bit machines.
     */

    sockaddrPtr->sin_addr.s_addr = addr.s_addr;
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpClient --
 *
 *	Opens a TCP client socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed.  An error message is returned
 *	in the interpreter on failure.
 *
 * Side effects:
 *	Opens a client socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpClient(interp, port, host, myaddr, myport)
    Tcl_Interp *interp;			/* For error reporting; can be NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Host on which to open port. */
    char *myaddr;			/* Client-side address */
    int myport;				/* Client-side port */
{
    Tcl_Channel chan;
    TcpState *statePtr;
    char channelName[20];

    /*
     * Create a new client socket and wrap it in a channel.
     */

    statePtr = CreateSocket(interp, port, host, 0, myaddr, myport);
    if (statePtr == NULL) {
	return NULL;
    }

    statePtr->acceptProc = NULL;
    statePtr->acceptProcData = (ClientData) NULL;

    sprintf(channelName, "sock%d",
	    (int) Tcl_GetFileInfo(statePtr->sock, NULL));

    chan = Tcl_CreateChannel(&tcpChannelType, channelName, statePtr->sock,
	    statePtr->sock, (ClientData) statePtr);
    if (chan != (Tcl_Channel) NULL) {
        Tcl_SetDefaultTranslation(chan, TCL_TRANSLATE_CRLF);
    }

    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_MakeTcpClientChannel --
 *
 *	Creates a Tcl_Channel from an existing client TCP socket.
 *
 * Results:
 *	The Tcl_Channel wrapped around the preexisting TCP socket.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_MakeTcpClientChannel(sock)
    ClientData sock;		/* The socket to wrap up into a channel. */
{
    TcpState *statePtr;
    Tcl_File sockFile;
    char channelName[20];
    Tcl_Channel chan;

    sockFile = Tcl_GetFile(sock, TCL_VMS_SOCKET);
    statePtr = (TcpState *) ckalloc((unsigned) sizeof(TcpState));
    statePtr->sock = sockFile;
    statePtr->acceptProc = NULL;
    statePtr->acceptProcData = (ClientData) NULL;
    statePtr->chan = (int)sock;
    statePtr->eventFlag = 0;
    statePtr->checkMask = 0;
    statePtr->watchMask = 0;
    Tcl_SetNotifierData(statePtr->sock, socketFreeProc, (ClientData)statePtr);

    sprintf(channelName, "sock%d", (int) sock);
    
    chan = Tcl_CreateChannel(&tcpChannelType, channelName, sockFile, sockFile,
            (ClientData) statePtr);
    if (chan != (Tcl_Channel) NULL) {
        Tcl_SetDefaultTranslation(chan, TCL_TRANSLATE_CRLF);
    }
    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpServer --
 *
 *	Opens a TCP server socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed.
 *
 * Side effects:
 *	Opens a server socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpServer(interp, port, host, acceptProc, acceptProcData)
    Tcl_Interp *interp;			/* For error reporting - may be
                                         * NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Name of local host. */
    Tcl_TcpAcceptProc *acceptProc;	/* Callback for accepting connections
                                         * from new clients. */
    ClientData acceptProcData;		/* Data for the callback. */
{
    Tcl_Channel chan;
    TcpState *statePtr;
    char channelName[20];

    /*
     * Create a new client socket and wrap it in a channel.
     */

    statePtr = CreateSocket(interp, port, host, 1, NULL, 0);
    if (statePtr == NULL) {
	return NULL;
    }

    statePtr->acceptProc = acceptProc;
    statePtr->acceptProcData = acceptProcData;

    /*
     * Set up the callback mechanism for accepting connections
     * from new clients. The caller will use Tcl_TcpRegisterCallback
     * to register a callback to call when a new connection is
     * accepted.
     */

    Tcl_CreateFileHandler(statePtr->sock, TCL_READABLE, TcpAccept,
            (ClientData) statePtr);
    
    sprintf(channelName, "sock%d",
	    (int) Tcl_GetFileInfo(statePtr->sock, NULL));

    chan = Tcl_CreateChannel(&tcpChannelType, channelName, NULL, NULL,
            (ClientData) statePtr);
    if (chan != (Tcl_Channel) NULL) {
        Tcl_SetDefaultTranslation(chan, TCL_TRANSLATE_CRLF);
    }

    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpAccept --
 *	Accept a TCP socket connection.  This is called by the event loop,
 *	and it in turns calls any registered callbacks for this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Evals the Tcl script associated with the server socket.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
TcpAccept(data, mask)
    ClientData data;			/* Callback token. */
    int mask;				/* Not used. */
{
    TcpState *sockState;		/* Client data of server socket. */
    int newsock;			/* The new client socket */
    Tcl_File newFile;			/* Its file. */
    TcpState *newSockState;		/* State for new socket. */
    struct sockaddr_in addr;		/* The remote address */
    int len;				/* For accept interface */
    Tcl_Channel chan;			/* Channel instance created. */
    char channelName[20];

    sockState = (TcpState *) data;

    len = sizeof(struct sockaddr_in);
    newsock = accept((int) Tcl_GetFileInfo(sockState->sock, NULL),
	    (struct sockaddr *)&addr, (unsigned int *)&len);
    if (newsock < 0) {
        return;
    }
    
    newFile = Tcl_GetFile((ClientData) newsock, TCL_VMS_SOCKET);
    if (newFile) {
        newSockState = (TcpState *) ckalloc((unsigned) sizeof(TcpState));

        newSockState->flags = 0;
        newSockState->sock = newFile;
        newSockState->acceptProc = (Tcl_TcpAcceptProc *) NULL;
        newSockState->acceptProcData = (ClientData) NULL;
	newSockState->chan = newsock;
	newSockState->eventFlag = 0;
	newSockState->checkMask = 0;
	newSockState->watchMask = 0;
        
        sprintf(channelName, "sock%d", (int) newsock);
        chan = Tcl_CreateChannel(&tcpChannelType, channelName, newFile,
                newFile, (ClientData) newSockState);
        if (chan == (Tcl_Channel) NULL) {
	    if (newSockState->eventFlag)
		freeEventFlag(newSockState->eventFlag);
            ckfree((char *) newSockState);
            socket_close(newsock);
            Tcl_FreeFile(newFile);
	} else {
            Tcl_SetDefaultTranslation(chan, TCL_TRANSLATE_CRLF);
            if (sockState->acceptProc != (Tcl_TcpAcceptProc *) NULL) {
                (sockState->acceptProc) (sockState->acceptProcData, chan,
                        inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
            }
	}
    }
}

static void
socketFreeProc(ClientData clientdata)
{
    /*
     * Free up socket info structure
     */
    TcpState *statePtr = (TcpState *)clientdata;

    if (statePtr) {
	if (statePtr->eventFlag)
	    freeEventFlag(statePtr->eventFlag);
	ckfree((char *)statePtr);
    }
}

#define	SELECT_DONTWAIT		(1<<0)
#define	SELECT_READABLE		(1<<1)
#define	SELECT_WRITEABLE	(1<<2)
#define	SELECT_EXCEPTION	(1<<3)

TclVmsWatchSocket(Tcl_File file, int mask)
{
    /*
     * Watch for something to happen on a socket.
     *
     * Start a QIO and return the event flag
     * that will be set when something happens.
     */
    int flag, status;
    TcpState *statePtr = (TcpState *)Tcl_GetNotifierData(file, NULL);

    if ((flag = statePtr->eventFlag) == 0) {
	flag = statePtr->eventFlag = getEventFlag();
    }
    statePtr->watchMask = 0;
    if (mask & TCL_READABLE)
	statePtr->watchMask |= SELECT_READABLE;
    if (mask & TCL_WRITABLE)
	statePtr->watchMask |= SELECT_WRITEABLE;
    if (mask & TCL_EXCEPTION)
	statePtr->watchMask |= SELECT_EXCEPTION;
    status = sys$qio(flag, statePtr->chan, IO$_SELECT, statePtr->iosb,
		     socketReadyHandler, statePtr,
		     &statePtr->watchMask,
		     0, 0, 0, 0, 0);
    if (status & 1)
	statePtr->flags |= TCP_WATCHING_SOCKET;
    else
	printf("IO$SELECT failed: %x\n", status);
    return flag;
}

VmsSocketReady(Tcl_File file, int mask)
{
    /*
     * Check and see what's ready for a Multinet socket
     */
    int readyMask = 0;
    TcpState *statePtr = (TcpState *)Tcl_GetNotifierData(file, NULL);

    /*
     *Convert Multinet flags to TCL flags
     */
    if (statePtr->watchMask & SELECT_READABLE)
	readyMask |= TCL_READABLE;
    if (statePtr->watchMask & SELECT_WRITEABLE)
	readyMask |= TCL_WRITABLE;
    if (statePtr->watchMask & SELECT_EXCEPTION)
	readyMask |= TCL_EXCEPTION;
    return (readyMask & mask);
}

static void
socketReadyHandler(TcpState *statePtr)
{
    /*
     * AST routine called when something is ready on a socket.
     */
    if (statePtr->iosb[0] & 1)
	statePtr->flags &= ~TCP_WATCHING_SOCKET;
    else
	printf("IO$SELECT failed(AST): %x\n", statePtr->iosb[1]);
}
